#include <string.h>
#include <stdlib.h>
#include <assert.h>
#include <errno.h>

#include "config.h"
#include "passwd.hpp"

CPasswdEntry::CPasswdEntry(FILE *fpFile) : CEntry(fpFile, (void *)&strEntry, sizeof(struct passwd), &GetEntry)
{
}

CPasswdEntry::CPasswdEntry(char *pszName,
									uid_t i,
									gid_t j,
									char *pszGecos,
									char *pszDirectory,
									char *pszShell)
{
	iErrorCode = NO_ERROR;
	
	strEntry.pw_name = strdup(pszName);
#ifdef SHADOWPWD	
	strEntry.pw_passwd = strdup("x");
#else	
	strEntry.pw_passwd = strdup("*");
#endif // SHADOW_PWD
	strEntry.pw_uid = i;
	strEntry.pw_gid = j;
#ifdef ATT_AGE
	strEntry.pw_age = strdup("");
#endif // ATT_AGE
#ifdef ATT_COMMENT
	strEntry.pw_comment = strdup("");
#endif // ATT_COMMENT
	strEntry.pw_gecos = strdup(pszGecos);
	strEntry.pw_dir = strdup(pszDirectory);
	strEntry.pw_shell = strdup(pszShell);
	if (NULL == strEntry.pw_name ||
		 NULL == strEntry.pw_passwd ||
#ifdef ATT_AGE
		 NULL == strEntry.pw_age ||
#endif // ATT_AGE
#ifdef ATT_COMMENT
		 NULL == strEntry.pw_comment ||
#endif // ATT_COMMENT
		 NULL == strEntry.pw_gecos ||
		 NULL == strEntry.pw_dir ||
		 NULL == strEntry.pw_shell)
		iErrorCode = MEMORY_ERROR;
}

CPasswdEntry::~CPasswdEntry()
{
	free(strEntry.pw_name);
	free(strEntry.pw_passwd);
#ifdef ATT_AGE
	free(strEntry.pw_age);
#endif // ATT_AGE
#ifdef ATT_COMMENT
	free(strEntry.pw_comment);
#endif // ATT_COMMENT
	free(strEntry.pw_gecos);
	free(strEntry.pw_dir);
	free(strEntry.pw_shell);
}

char *CPasswdEntry::GetName()
{
	return strEntry.pw_name;
}

char *CPasswdEntry::GetPassword()
{
	return strEntry.pw_passwd;
}

gid_t CPasswdEntry::GetGroup()
{
	return strEntry.pw_gid;
}

uid_t CPasswdEntry::GetUser()
{
	return strEntry.pw_uid;
}

bool CPasswdEntry::SetPassword(char *pszPassword, int, int, int, int)
{
	char *pszNewPassword = strdup(pszPassword);
	if (NULL == pszNewPassword)
		return false;
	free(strEntry.pw_passwd);
	strEntry.pw_passwd = pszNewPassword;
	return true;
}

bool CPasswdEntry::GetEntry(char *pszEntry, void *pstrTarget)
{
	assert(NULL != pszEntry);
	assert(NULL != pstrTarget);
	
	struct passwd *pstrEntry = (struct passwd *)pstrTarget;
	
	register int i, j;
	
	for (i = j = 0; pszEntry[i] != '\0'; i++)
		if (pszEntry[i] == ':')
	  		j += 1;
	
	if (6 != j)
		return false;
	
	bool fgError = false;
	pstrEntry->pw_name = strdup(strsep(&pszEntry, ":"));
	pstrEntry->pw_passwd = strdup(strsep(&pszEntry, ":"));
	pstrEntry->pw_uid = ToLong(strsep(&pszEntry, ":"), -1, &fgError);
	pstrEntry->pw_gid = ToLong(strsep(&pszEntry, ":"), -1, &fgError);
	pstrEntry->pw_gecos = strdup(strsep(&pszEntry, ":"));
	pstrEntry->pw_dir = strdup(strsep(&pszEntry, ":"));
	pstrEntry->pw_shell = strdup(strsep(&pszEntry, ":"));	
	
	if ((NULL == pstrEntry->pw_name) || 
		 (NULL == pstrEntry->pw_passwd) ||
		 (NULL == pstrEntry->pw_gecos) ||
		 (NULL == pstrEntry->pw_dir) ||
		 (NULL == pstrEntry->pw_shell))
		return false;
	
#ifdef ATT_AGE
	pstrEntry->pw_age = pstrEntry->pw_passwd;
	while((',' != pstrEntry->pw_age) && ('\0' != pstrEntry->pw_age))
		pstrEntry->pw_age += 1;
#endif // ATT_AGE	
#ifdef ATT_COMMENT
	pstrEntry->pw_comment = "";
#endif // ATT_COMMENT
	
	return true;
}

bool CPasswdEntry::PutEntry(FILE *fpFile)
{
	if ((NULL == strEntry.pw_name) || 
		 (NULL == strEntry.pw_passwd) ||
		 (NULL == strEntry.pw_gecos) ||
		 (NULL == strEntry.pw_dir) ||
		 (NULL == strEntry.pw_shell))
		return false;
	
	fprintf(fpFile,
			  "%s:%s"
#ifdef ATT_AGE				
			  ",%s"
#endif // ATT_AGE				
			  ":%d:%d:%s"
#ifdef ATT_COMMENT				
			  ",%s"
#endif // ATT_COMMENT				
			  ":%s:%s\n",
			  strEntry.pw_name,
			  strEntry.pw_passwd,
#ifdef ATT_AGE				
			  strEntry.pw_age,
#endif // ATT_AGE				
			  strEntry.pw_uid,
			  strEntry.pw_gid,
			  strEntry.pw_gecos,
#ifdef ATT_COMMENT				
			  strEntry.pw_comment,
#endif // ATT_COMMENT				
			  strEntry.pw_dir,
			  strEntry.pw_shell);
	
	return true;
}