#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <syslog.h>
#include <string.h>
#include <arpa/inet.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include <errno.h>

#include "globals.h"
#include "getdef.h"
#include "sockio.h"
#include "conf.hpp"

#define SERV_TCP_PORT 1099
#define CONF_FILE		 "/etc/passwdc.conf"

#define strzero(p) memset(p, 0, strlen(p))
#define memzero(p) memset(p, 0, sizeof(p))

int iMax = 99999;
int iMin = 0;
int iWarn = -1;
int iInact = -1;

// Function prototypes
int option_value(char *, bool *, int, int, char *);

int main(int argc, char **argv)
{
	static char szServers[512];
	static char szBuf[128];
	
	static char szPassword[128];
	static char szNewPassword1[128];
	static char szNewPassword2[128];
	
	struct sockaddr_in client;
	struct hostent *pHost;
	
	char *pszLogin = (char *)NULL;
	char *pszTarget = (char *)NULL;
	char *pszServer;
	char *pszPassword;
	char *pszNewPassword1;
	char *pszNewPassword2;
	
	int sockfd;
	
	int iMaxLen;
	int iMinLen;
	
	int iServerPort;
	
	int iResult = 0;
	
	int iParams = 0;
	
	char *pszConfig = getenv("PASSWDCCONF");
	char *pszValue = NULL;
	bool fgError = false;

	if (argc > 1)
		for (int i = 1; i < argc; i++)
			if ('-' == argv[i][0]) {				// We have option
				switch (argv[i][1]) {
					case 'c':							// We have other config file
						if ('\0' != argv[i][2]) {	// The user has not typed space after -c
							pszConfig = &argv[i][2];
						} else {
							pszConfig = i < (argc - 1) ? argv[i + 1] : (char *)NULL;
							if (NULL == pszConfig) {
								fprintf(stderr, "Invalid configuration file specified...\n");
								fgError = true;
							}
							i += 1;
						}
					break;
					case 'x':
						if ('\0' != argv[i][2]) {	// The user has not typed space after -c
							pszValue = &argv[i][2];
						} else {
							pszValue = i < (argc - 1) ? argv[i + 1] : (char *)NULL;
							if (NULL == pszValue) {
								fprintf(stderr, "Invalid value for option: -x...\n");
								fgError = true;
							}
						}
						if (NULL != pszValue)
							iMax = option_value(pszValue, &fgError, 0, 99999, "-x");
					break;
					case 'n':
						if ('\0' != argv[i][2]) {	// The user has not typed space after -c
							pszValue = &argv[i][2];
						} else {
							pszValue = i < (argc - 1) ? argv[i + 1] : (char *)NULL;
							if (NULL == pszValue) {
								fprintf(stderr, "Invalid value for option: -n...\n");
								fgError = true;
							}
						}
						if (NULL != pszValue)
							iMin = option_value(pszValue, &fgError, 0, 99999, "-n");
					break;
					case 'w':
						if ('\0' != argv[i][2]) {	// The user has not typed space after -c
							pszValue = &argv[i][2];
						} else {
							pszValue = i < (argc - 1) ? argv[i + 1] : (char *)NULL;
							if (NULL == pszValue) {
								fprintf(stderr, "Invalid value for option: -w...\n");
								fgError = true;
							}
						}
						if (NULL != pszValue)
							iWarn = option_value(pszValue, &fgError, 0, 99999, "-w");
					break;
					case 'i':
						if ('\0' != argv[i][2]) {	// The user has not typed space after -c
							pszValue = &argv[i][2];
						} else {
							pszValue = i < (argc - 1) ? argv[i + 1] : (char *)NULL;
							if (NULL == pszValue) {
								fprintf(stderr, "Invalid value for option: -i...\n");
								fgError = true;
							}
						}
						if (NULL != pszValue)
							iInact = option_value(pszValue, &fgError, 0, 99999, "-i");
					break;
					default:
						fprintf(stderr, "Invalid option: %s...\n", argv[i]);
						fgError = true;
					break;
				}
			} else {
				if (0 == iParams)
					pszTarget = argv[i];
				else if (1 == iParams)
				  pszLogin = argv[i];
				else
					fgError = true;
				iParams += 1;
			}

	if ((argc < 1) || (argc > 13) || (true == fgError)) {
		printf("Usage: passwdc [<target>] [<user>] [-c <config file>]\n");
		printf("       [-x <max>] [-n <min>] [-w <warn>] [-i <inact>]\n");
		printf("Where: <target>     is the user whose password you want to change;\n");
		printf("       <login>      is your login if you are master user for the given target.\n");
		printf("Options: -c <config file>\n");
		printf("                    use alternate config file instead the default\n");
		printf("	                   '/etc/passwdc.conf' system wide config file;\n");
		printf("         -x <max>   sets the maximum number of days a password remains valid;\n");
		printf("         -n <min>   sets the minimum number of days before a password may be\n");
		printf("                    changed;\n");
		printf("         -w <warn>  sets the number of days before a password expiration a\n");
		printf("                    user will receive a warning;\n");
		printf("         -i <inact> sets the number of days after password expiration, the\n");
		printf("                    account will be marked inactive.\n");
		printf("Written by: <Alexander Feldman> alex@varna.net\n\n");
		return 1;
	}
	
	if (NULL == pszConfig)
		pszConfig = CONF_FILE;
	
	char *pszMessage = NULL;	
	CConfigurationFile *pConfigFile = new CConfigurationFile(pszConfig, &pszMessage);
	if (NULL == pConfigFile) {
		fprintf(stderr, "Error allocating memory...\n");
		return 1;
	}
	if (NULL != pszMessage) {
		fprintf(stderr, "%s\n", pszMessage);
		delete pConfigFile;
		return 1;
	}
	
	iServerPort = pConfigFile->GetInteger("Global", "port", SERV_TCP_PORT);
	memzero(szServers);
	strncpy(szServers, pConfigFile->GetString("Global", "targets", "localhost"), sizeof(szServers) - 1);
	
	openlog(argv[0], LOG_PID, LOG_USER);

	if (NULL == pszLogin)
		pszLogin = getlogin();
	if (NULL == pszTarget)
		pszTarget = pszLogin;
	
	pszPassword = getpass("Old password:");
	strcpy(szPassword, pszPassword);
	strzero(pszPassword);

// In fact we don't need to check for login lengths on the source machine but
// on the target but it always better to have sample restrictions for the
// min and max lengths of a string than not to have.	
	iMinLen = getdef_num("PASS_MIN_LEN", 5);
	iMaxLen = getdef_num("PASS_MAX_LEN", 8);
	printf("Enter the new password (minimum of %d, maximum of %d characters)\n", iMinLen, iMaxLen);
	printf("Please use combination of upper and lower case letters and numbers.\n");
	
	pszNewPassword1 = getpass("New password:");
	strcpy(szNewPassword1, pszNewPassword1);	
	strzero(pszNewPassword1);
	pszNewPassword2 = getpass("Re-enter new password:");
	strcpy(szNewPassword2, pszNewPassword2);
	strzero(pszNewPassword2);
	
// We can skip overflow and error checking from getpass(), because in all
// cases it returns valid pointer and the internal buffer is 128 bytes and
// null terminated.	
	
	if (0 != strcmp(szNewPassword1, szNewPassword2)) {
		printf("They don't match; try again.\n");
		iResult = 1;
		goto err_exit;
	}
	if ((int)strlen(szNewPassword1) < iMinLen || 
		 (int)strlen(szNewPassword1) > iMaxLen) {
		printf("They don't meet length criteria; try again.\n");		
		iResult = 2;
		goto err_exit;
	}
	printf("\n");

	pszServer = strtok(szServers, ";,");
	do {
		pszServer += strspn(pszServer, " \t");			
		printf("Changing password for %s on %s\n", pszTarget, pszServer);
	
		if ((sockfd = socket(PF_INET, SOCK_STREAM, 0)) < 0) {
			printf("Socket error: %s\n", strerror(errno));
			continue;
		}

		pHost = gethostbyname(pszServer);
		if (NULL == pHost) {
			printf("Resolving error: %s\n", pszServer);
			close(sockfd);
			continue;
		}
		memset(&client, 0, sizeof(client));
		client.sin_family = AF_INET;
		client.sin_addr.s_addr = *((unsigned long int *)(pHost->h_addr));
		client.sin_port = htons(iServerPort);
		if (connect(sockfd, (struct sockaddr *)&client, sizeof(client)) < 0) {
			printf("Connect error: %s\n", strerror(errno));
			close(sockfd);
			continue;
		}

		unsigned char c = '\0';
		read_byte(sockfd, &c);
		if (WELCOME != c) {
			printf("Invalid response from server.\n");
			close(sockfd);
			continue;
		} 
		if (((int)strlen(pszLogin) + 1 != write_string(sockfd, pszLogin)) ||
			 ((int)strlen(szPassword) + 1 != write_string(sockfd, szPassword))) {
			printf("Write error.\n");
			close(sockfd);
			continue;
		} 
		read_byte(sockfd, &c);			
		switch (c) {
			case GO_AHEAD:
				write_byte(sockfd, ACTION_CHANGEPASSWORD);
				if (((int)strlen(pszTarget) + 1 != write_string(sockfd, pszTarget)) ||
					 ((int)strlen(szNewPassword1) + 1 != write_string(sockfd, szNewPassword1)) ||
					 (false == write_integer(sockfd, iMin)) ||
					 (false == write_integer(sockfd, iMax)) ||
					 (false == write_integer(sockfd, iWarn)) ||
					 (false == write_integer(sockfd, iInact)))
					printf("Write error.\n");
				read_byte(sockfd, &c);
				switch (c) {
					case OK:
						printf("Password successfully changed.\n");
					break;
					case INVALID_USER:
						printf("Invalid target specified.\n");
					break;
					case ERROR:
						printf("Password still unchanged. Server returned error.\n");
					break;
					case NO_RIGHTS:
						printf("Access denied.\n");
					break;
					case WEAK_PASSWORD:
						read_string(sockfd, szBuf, sizeof(szBuf));
						printf("Weak password: %s.\n", szBuf);
					break;
					default:
						printf("Invalid response from server.\n");
				}
			break;
			case LOGIN_FAILED:
				printf("Incorrect login or password for %s on %s.\n", pszLogin, pszServer);				
			break;
			default:
				printf("<%c>Invalid response from server.\n", c);
		}
		close(sockfd);
	} while (NULL != (pszServer = strtok(NULL, ";,")));

err_exit:
	closelog();

	strzero(szPassword);
	strzero(szNewPassword1);
	strzero(szNewPassword2);

	delete pConfigFile;
	
	return iResult;
}

int option_value(char *pszValue, bool *pfgError, int iMin, int iMax, char *pszInfo)
{
	int iResult = 0;
	
	if ((NULL == pszValue) || ('\0' == *pszValue)) {
		fprintf(stderr, "Invalid value for option: %s...\n", pszInfo);
		*pfgError = true;
		return iResult;
	}
	char *pszInvalid;
	iResult = strtol(pszValue, &pszInvalid, 10);
	if ('\0' != *pszInvalid) {
		fprintf(stderr, "Invalid value for option: %s...\n", pszInfo);
		*pfgError = true;
	}
	if (iResult < iMin || iResult > iMax) {
		fprintf(stderr, "Out of range value for option: %s...\n", pszInfo);
		*pfgError = true;
	}
	
	return iResult;
}